"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const path_1 = require("path");
const fs_1 = require("fs");
const knex = require("knex");
const model_1 = require("./model");
const helpers_1 = require("./helpers");
const schema_helpers_1 = require("./schema-helpers");
const sqljs_handler_1 = require("./sqljs-handler");
const util_1 = require("./util");
const types = require("./types");
const ensureExists = (atPath) => {
    try {
        fs_1.closeSync(fs_1.openSync(atPath, 'wx'));
    }
    catch (_a) { }
};
class Trilogy {
    constructor(path, options = {}) {
        this.Model = model_1.default;
        util_1.invariant(path, 'trilogy constructor must be provided a file path');
        const obj = this.options =
            types.validate(options, types.TrilogyOptions);
        if (path === ':memory:') {
            obj.connection.filename = path;
        }
        else {
            obj.connection.filename = path_1.resolve(obj.dir, path);
            // ensure the directory exists
            util_1.makeDirPath(path_1.dirname(obj.connection.filename));
        }
        this.isNative = obj.client === 'sqlite3';
        this.verbose = obj.verbose;
        const config = { client: 'sqlite3', useNullAsDefault: true };
        if (this.isNative) {
            if (path !== ':memory:') {
                ensureExists(obj.connection.filename);
            }
            this.knex = knex(Object.assign({}, config, { connection: obj.connection }));
        }
        else {
            this.knex = knex(config);
            this.pool = sqljs_handler_1.pureConnect(this);
            sqljs_handler_1.readDatabase(this);
        }
        this._definitions = new Map();
    }
    static connect(path, options = {}) {
        return new this(path, options);
    }
    get models() {
        return [...this._definitions.keys()];
    }
    async model(name, schema, options = {}) {
        if (this._definitions.has(name)) {
            return this._definitions.get(name);
        }
        const model = new this.Model(this, name, schema, options);
        this._definitions.set(name, model);
        const opts = schema_helpers_1.toKnexSchema(model, types.validate(options, types.ModelOptions, {}));
        const check = this.knex.schema.hasTable(name);
        const query = this.knex.schema.createTable(name, opts);
        if (this.isNative) {
            // tslint:disable-next-line:await-promise
            const exists = await check;
            return exists ? model : query.then(() => model);
        }
        else {
            const exists = await helpers_1.runQuery(this, check, true);
            if (exists)
                return model;
            return helpers_1.runQuery(this, query).then(() => model);
        }
    }
    getModel(name) {
        return util_1.invariant(this._definitions.get(name), `no model defined by the name '${name}'`);
    }
    async hasModel(name) {
        if (!this._definitions.has(name)) {
            return false;
        }
        const query = this.knex.schema.hasTable(name);
        return helpers_1.runQuery(this, query, true);
    }
    async dropModel(name) {
        if (!this._definitions.has(name)) {
            return false;
        }
        const query = this.knex.schema.dropTableIfExists(name);
        await helpers_1.runQuery(this, query, true);
        this._definitions.delete(name);
        return true;
    }
    raw(query, needResponse) {
        return helpers_1.runQuery(this, query, needResponse);
    }
    close() {
        if (this.isNative) {
            // must wrap this return value in native Promise due to
            // https://github.com/petkaantonov/bluebird/issues/1277
            return Promise.resolve(this.knex.destroy());
        }
        else {
            return this.pool.drain();
        }
    }
    create(table, object, options) {
        const model = this.getModel(table);
        return model.create(object, options);
    }
    find(location, criteria, options) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        if (column) {
            return model.findIn(column, criteria, options);
        }
        else {
            return model.find(criteria, options);
        }
    }
    findOne(location, criteria, options) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        if (column) {
            return model.findOneIn(column, criteria, options);
        }
        else {
            return model.findOne(criteria, options);
        }
    }
    findOrCreate(table, criteria, creation, options) {
        const model = this.getModel(table);
        return model.findOrCreate(criteria, creation, options);
    }
    update(table, criteria, data, options) {
        const model = this.getModel(table);
        return model.update(criteria, data, options);
    }
    updateOrCreate(table, criteria, data, options) {
        const model = this.getModel(table);
        return model.updateOrCreate(criteria, data, options);
    }
    get(location, criteria, defaultValue) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.get(column, criteria, defaultValue);
    }
    set(location, criteria, value) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.set(column, criteria, value);
    }
    getRaw(location, criteria, defaultValue) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.getRaw(column, criteria, defaultValue);
    }
    setRaw(location, criteria, value) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.setRaw(column, criteria, value);
    }
    incr(location, criteria, amount) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.incr(column, criteria, amount);
    }
    decr(location, criteria, amount, allowNegative) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.decr(column, criteria, amount, allowNegative);
    }
    remove(location, criteria) {
        const model = this.getModel(location);
        return model.remove(criteria);
    }
    clear(location) {
        const model = this.getModel(location);
        return model.clear();
    }
    count(location, criteria, options) {
        if (location == null && criteria == null && options == null) {
            const query = this.knex('sqlite_master')
                .whereNot('name', 'sqlite_sequence')
                .where({ type: 'table' })
                .count('* as count');
            return helpers_1.runQuery(this, query, true)
                .then(([{ count }]) => count);
        }
        const [table, column] = util_1.defaultTo(location, '').split('.', 2);
        const model = this.getModel(table);
        return column
            ? model.countIn(column, criteria, options)
            : model.count(criteria, options);
    }
    min(location, criteria, options) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.min(column, criteria, options);
    }
    max(location, criteria, options) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.max(column, criteria, options);
    }
}
exports.Trilogy = Trilogy;
var model_2 = require("./model");
exports.Model = model_2.default;
tslib_1.__exportStar(require("./types"), exports);
tslib_1.__exportStar(require("./plugins"), exports);
exports.connect = (path, options) => new Trilogy(path, options);
