"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const helpers = require("./helpers");
const schema_helpers_1 = require("./schema-helpers");
const util_1 = require("./util");
const types = require("./types");
exports.MODEL_FLAG = Symbol('trilogy-model');
class Model {
    constructor(ctx, name, schema, options) {
        this.ctx = ctx;
        this.name = name;
        this.options = options;
        this.schema = schema_helpers_1.normalizeSchema(schema);
        this.cast = new schema_helpers_1.Cast(this);
    }
    static get [exports.MODEL_FLAG]() {
        return true;
    }
    static [Symbol.hasInstance](object) {
        return object[exports.MODEL_FLAG] === true;
    }
    create(object, options = {}) {
        const insertion = this.cast.toDefinition(object, options);
        const query = this.ctx.knex.raw(this.ctx.knex(this.name)
            .insert(insertion)
            .toString()
            .replace(/^insert/i, 'INSERT OR IGNORE'));
        return helpers.runQuery(this.ctx, query)
            .then(() => helpers.findLastObject(this, object))
            .then(res => {
            if (res)
                return res;
            return res == null ? undefined : this.findOne(object);
        });
    }
    async find(criteria, options = {}) {
        options = types.validate(options, types.FindOptions);
        const order = options.random ? 'random' : options.order;
        let query = this.ctx.knex(this.name).select();
        query = helpers.buildWhere(query, criteria);
        if (order)
            query = helpers.buildOrder(query, order);
        if (options.limit)
            query = query.limit(options.limit);
        if (options.skip)
            query = query.offset(options.skip);
        const response = await helpers.runQuery(this.ctx, query, true);
        if (!Array.isArray(response)) {
            return response ? [response] : [];
        }
        return response.map(object => {
            return this.cast.fromDefinition(object, options);
        });
    }
    async findIn(column, criteria, options) {
        const response = await this.find(criteria, options);
        return response.map(object => {
            return this.cast.fromColumnDefinition(column, object[column], options);
        });
    }
    async findOne(criteria, options = {}) {
        options = types.validate(options, types.FindOptions);
        const order = options.random ? 'random' : options.order;
        let query = this.ctx.knex(this.name).first();
        query = helpers.buildWhere(query, criteria);
        if (order)
            query = helpers.buildOrder(query, order);
        if (options.skip)
            query = query.offset(options.skip);
        const response = await helpers.runQuery(this.ctx, query, true);
        const result = Array.isArray(response) ? response[0] : response;
        if (util_1.isNil(result))
            return undefined;
        return this.cast.fromDefinition(result, options);
    }
    async findOneIn(column, criteria, options = {}) {
        return this.findOne(criteria, options)
            .then(object => object != null ? object[column] : undefined);
    }
    async findOrCreate(criteria, creation = {}, options) {
        const existing = await this.findOne(criteria, options);
        // must cast `criteria` & `creation` to any as a workaround
        // for not being able to spread a generic type, see:
        // https://github.com/Microsoft/TypeScript/issues/10727
        return existing || this.create(Object.assign({}, criteria, creation));
    }
    update(criteria = {}, data = {}, options = {}) {
        options = types.validate(options, types.UpdateOptions);
        if (Object.keys(data).length < 1)
            return Promise.resolve(0);
        const typedData = this.cast.toDefinition(data, options);
        const typedCriteria = this.cast.toDefinition(criteria, options);
        let query = this.ctx.knex(this.name).update(typedData);
        query = helpers.buildWhere(query, typedCriteria);
        return helpers.runQuery(this.ctx, query);
    }
    async updateOrCreate(criteria, data, options = {}) {
        const found = await this.find(criteria, options);
        if (!found || !found.length) {
            return this.create(Object.assign({}, criteria, data), options)
                .then(res => Number(!!res));
        }
        else {
            return this.update(criteria, data, options);
        }
    }
    get(column, criteria, defaultValue) {
        return baseGet(this, column, criteria, defaultValue);
    }
    set(column, criteria, value) {
        return baseSet(this, column, criteria, value);
    }
    getRaw(column, criteria, defaultValue) {
        return baseGet(this, column, criteria, defaultValue, { raw: true });
    }
    setRaw(column, criteria, value) {
        return baseSet(this, column, criteria, value, { raw: true });
    }
    async incr(column, criteria, amount) {
        const cast = Number(amount);
        if (Number.isNaN(cast))
            amount = 1;
        if (amount === 0)
            return 0;
        let query = this.ctx.knex(this.name).increment(column, amount);
        query = helpers.buildWhere(query, criteria);
        return helpers.runQuery(this.ctx, query);
    }
    async decr(column, criteria, amount, allowNegative) {
        const cast = Number(amount);
        if (Number.isNaN(cast))
            amount = 1;
        if (amount === 0)
            return 0;
        let query = this.ctx.knex(this.name);
        const raw = allowNegative
            ? `${column} - ${amount}`
            : `MAX(0, ${column} - ${amount})`;
        query = query.update({ [column]: this.ctx.knex.raw(raw) });
        query = helpers.buildWhere(query, criteria);
        return helpers.runQuery(this.ctx, query);
    }
    async remove(criteria) {
        if (!helpers.isValidWhere(criteria) ||
            (util_1.isObject(criteria) && !Object.keys(criteria).length))
            return 0;
        let query = this.ctx.knex(this.name).del();
        query = helpers.buildWhere(query, criteria);
        return helpers.runQuery(this.ctx, query);
    }
    clear() {
        const query = this.ctx.knex(this.name).truncate();
        return helpers.runQuery(this.ctx, query);
    }
    async count(criteria, options = {}) {
        return baseCount(this, '*', criteria, options);
    }
    async countIn(column, criteria, options = {}) {
        return baseCount(this, column, criteria, options);
    }
    async min(column, criteria, options = {}) {
        return baseMinMax(this, 'min', column, criteria, options);
    }
    async max(column, criteria, options) {
        return baseMinMax(this, 'max', column, criteria, options);
    }
}
exports.default = Model;
async function baseCount(model, column, criteria, options = {}) {
    util_1.invariant(column && util_1.isString(column), `invalid column: expected string, got ${typeof column}`);
    options = types.validate(options, types.AggregateOptions);
    const val = `${column} as count`;
    const builder = model.ctx.knex(model.name);
    let query = options.distinct ? builder.countDistinct(val) : builder.count(val);
    query = helpers.buildWhere(query, criteria);
    if (options.group)
        query = query.groupBy(util_1.toArray(options.group));
    const res = await helpers.runQuery(model.ctx, query, true);
    if (!Array.isArray(res))
        return 0;
    return res[0].count;
}
async function baseMinMax(model, method, column, criteria, options = {}) {
    options = types.validate(options, types.AggregateOptions);
    const val = `${column} as ${method}`;
    let query = model.ctx.knex(model.name)[method](val);
    query = helpers.buildWhere(query, criteria);
    if (options.group)
        query = query.groupBy(util_1.toArray(options.group));
    const res = await helpers.runQuery(model.ctx, query, true);
    if (!Array.isArray(res))
        return undefined;
    return res[0][method];
}
async function baseGet(model, column, criteria, defaultValue, options) {
    const data = await model.findOneIn(column, criteria, options);
    return util_1.defaultTo(data, defaultValue);
}
async function baseSet(model, column, criteria, value, options) {
    util_1.invariant(model.schema[column], `no column by the name '${column}' is defined in '${model.name}'`);
    return model.update(criteria, {
        [column]: value
    }, options);
}
