"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("./constants");
const helpers_1 = require("./helpers");
const util_1 = require("./util");
const types = require("./types");
function toKnexSchema(model, options) {
    return (table) => {
        // every property of `model.schema` is a column
        util_1.eachObj(model.schema, (descriptor, name) => {
            // each column's value is either its type or a descriptor
            const type = getDataType(descriptor);
            const partial = table[toKnexMethod(type)](name);
            if (util_1.isFunction(descriptor) || !util_1.isObject(descriptor))
                return;
            const props = types.validate(descriptor, types.ColumnDescriptor, {});
            if ('nullable' in props) {
                if ('notNullable' in props) {
                    util_1.invariant(false, `can't set both 'nullable' & 'notNullable' - they work inversely`);
                }
                props.notNullable = !props.nullable;
            }
            util_1.eachObj(props, (value, property) => {
                if (constants_1.IGNORABLE_PROPS.includes(property))
                    return;
                if (constants_1.KNEX_NO_ARGS.includes(property)) {
                    props[property] && partial[property]();
                }
                else {
                    props[property] && partial[property](value);
                }
            });
        });
        for (const key of Object.keys(options)) {
            const value = options[key];
            if (key === 'timestamps') {
                options.timestamps && table.timestamps(true, true);
            }
            else if (key === 'index') {
                createIndices(table, value);
            }
            else {
                table[key](value);
            }
        }
    };
}
exports.toKnexSchema = toKnexSchema;
function createIndices(table, value) {
    if (util_1.isString(value)) {
        table.index([value]);
    }
    else if (Array.isArray(value)) {
        if (value.every(util_1.isString)) {
            table.index(value);
        }
        value.forEach(columns => table.index(columns));
    }
    else if (util_1.isObject(value)) {
        util_1.eachObj(value, (columns, indexName) => {
            table.index(util_1.toArray(columns), indexName);
        });
    }
}
function castValue(value) {
    const type = typeof value;
    if (type === 'number' || type === 'string') {
        return value;
    }
    if (type === 'boolean')
        return Number(value);
    if (Array.isArray(value) || util_1.isObject(value)) {
        return JSON.stringify(value);
    }
    return value;
}
exports.castValue = castValue;
function normalizeSchema(schema) {
    const keys = Object.keys(schema);
    util_1.invariant(keys.length > 0, 'model schemas cannot be empty');
    const result = {};
    for (const key of keys) {
        const descriptor = schema[key];
        const type = typeof descriptor;
        result[key] = type === 'function' || type === 'string'
            ? { type: descriptor }
            : descriptor;
    }
    return result;
}
exports.normalizeSchema = normalizeSchema;
function getDataType(property) {
    let type = property;
    if (util_1.isFunction(property)) {
        type = property.name;
    }
    else if (util_1.isObject(property)) {
        type = util_1.isFunction(property.type)
            ? property.type.name
            : property.type;
    }
    if (util_1.isString(type)) {
        const lower = type.toLowerCase();
        if (!constants_1.COLUMN_TYPES.includes(lower)) {
            return 'string';
        }
        return lower;
    }
    return util_1.invariant(false, `column type must be of type string`);
}
function toKnexMethod(type) {
    switch (type) {
        case 'string':
        case 'array':
        case 'object':
        case 'json':
            return 'text';
        case 'number':
        case 'boolean':
            return 'integer';
        case 'date':
            return 'dateTime';
        case 'increments':
            return 'increments';
        default:
            return util_1.invariant(false, `invalid column type definition: ${type}`);
    }
}
exports.toKnexMethod = toKnexMethod;
function toInputType(type, value) {
    switch (type) {
        case 'string':
            return String(value);
        case 'array':
        case 'object':
        case 'json':
            return JSON.stringify(value);
        case 'number':
        case 'boolean':
        case 'increments':
            return Number(value);
        case 'date':
            return new Date(value);
        default:
            return util_1.invariant(false, `invalid type on insert to database: ${type}`);
    }
}
exports.toInputType = toInputType;
function toReturnType(type, value) {
    switch (type) {
        case 'string':
            return String(value);
        case 'array':
        case 'object':
        case 'json':
            return JSON.parse(value);
        case 'number':
        case 'increments':
            return Number(value);
        case 'boolean':
            return Boolean(value);
        case 'date':
            return new Date(value);
        default:
            return util_1.invariant(false, `invalid type returned from database: ${type}`);
    }
}
exports.toReturnType = toReturnType;
class Cast {
    constructor(model) {
        this.model = model;
    }
    toDefinition(object, options) {
        if (helpers_1.isWhereTuple(object)) {
            const clone = object.slice();
            const valueIndex = clone.length - 1;
            clone[valueIndex] =
                this.toColumnDefinition(clone[0], clone[valueIndex], options);
            return clone;
        }
        if (helpers_1.isWhereMultiple(object)) {
            return object.map(clause => this.toDefinition(clause, options));
        }
        if (util_1.isObject(object)) {
            return util_1.mapObj(object, (value, column) => {
                return this.toColumnDefinition(column, value, options);
            });
        }
        return util_1.invariant(false, `invalid input type: '${typeof object}'`);
    }
    fromDefinition(object, options) {
        return util_1.mapObj(object, (value, column) => {
            return this.fromColumnDefinition(column, value, options);
        });
    }
    toColumnDefinition(column, value, options = { raw: false }) {
        const definition = this.model.schema[column];
        util_1.invariant(!(definition.notNullable && value == null), `${this.model.name}.${column} is not nullable but received nil`);
        const type = getDataType(definition);
        const cast = value !== null ? toInputType(type, value) : value;
        if (!options.raw && util_1.isFunction(definition.set)) {
            return castValue(definition.set(cast));
        }
        return cast;
    }
    fromColumnDefinition(column, value, options = { raw: false }) {
        const definition = this.model.schema[column];
        const type = getDataType(definition);
        const cast = value !== null ? toReturnType(type, value) : value;
        if (!options.raw && util_1.isFunction(definition.get)) {
            return definition.get(cast);
        }
        return cast;
    }
}
exports.Cast = Cast;
