import { dirname, resolve } from 'path';
import { openSync, closeSync } from 'fs';
import * as knex from 'knex';
import Model from './model';
import { runQuery } from './helpers';
import { toKnexSchema } from './schema-helpers';
import { pureConnect, readDatabase } from './sqljs-handler';
import { defaultTo, invariant, makeDirPath } from './util';
import * as types from './types';
const ensureExists = (atPath) => {
    try {
        closeSync(openSync(atPath, 'wx'));
    }
    catch (_a) { }
};
export class Trilogy {
    constructor(path, options = {}) {
        this.Model = Model;
        invariant(path, 'trilogy constructor must be provided a file path');
        const obj = this.options =
            types.validate(options, types.TrilogyOptions);
        if (path === ':memory:') {
            obj.connection.filename = path;
        }
        else {
            obj.connection.filename = resolve(obj.dir, path);
            // ensure the directory exists
            makeDirPath(dirname(obj.connection.filename));
        }
        this.isNative = obj.client === 'sqlite3';
        this.verbose = obj.verbose;
        const config = { client: 'sqlite3', useNullAsDefault: true };
        if (this.isNative) {
            if (path !== ':memory:') {
                ensureExists(obj.connection.filename);
            }
            this.knex = knex(Object.assign({}, config, { connection: obj.connection }));
        }
        else {
            this.knex = knex(config);
            this.pool = pureConnect(this);
            readDatabase(this);
        }
        this._definitions = new Map();
    }
    static connect(path, options = {}) {
        return new this(path, options);
    }
    get models() {
        return [...this._definitions.keys()];
    }
    async model(name, schema, options = {}) {
        if (this._definitions.has(name)) {
            return this._definitions.get(name);
        }
        const model = new this.Model(this, name, schema, options);
        this._definitions.set(name, model);
        const opts = toKnexSchema(model, types.validate(options, types.ModelOptions, {}));
        const check = this.knex.schema.hasTable(name);
        const query = this.knex.schema.createTable(name, opts);
        if (this.isNative) {
            // tslint:disable-next-line:await-promise
            const exists = await check;
            return exists ? model : query.then(() => model);
        }
        else {
            const exists = await runQuery(this, check, true);
            if (exists)
                return model;
            return runQuery(this, query).then(() => model);
        }
    }
    getModel(name) {
        return invariant(this._definitions.get(name), `no model defined by the name '${name}'`);
    }
    async hasModel(name) {
        if (!this._definitions.has(name)) {
            return false;
        }
        const query = this.knex.schema.hasTable(name);
        return runQuery(this, query, true);
    }
    async dropModel(name) {
        if (!this._definitions.has(name)) {
            return false;
        }
        const query = this.knex.schema.dropTableIfExists(name);
        await runQuery(this, query, true);
        this._definitions.delete(name);
        return true;
    }
    raw(query, needResponse) {
        return runQuery(this, query, needResponse);
    }
    close() {
        if (this.isNative) {
            // must wrap this return value in native Promise due to
            // https://github.com/petkaantonov/bluebird/issues/1277
            return Promise.resolve(this.knex.destroy());
        }
        else {
            return this.pool.drain();
        }
    }
    create(table, object, options) {
        const model = this.getModel(table);
        return model.create(object, options);
    }
    find(location, criteria, options) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        if (column) {
            return model.findIn(column, criteria, options);
        }
        else {
            return model.find(criteria, options);
        }
    }
    findOne(location, criteria, options) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        if (column) {
            return model.findOneIn(column, criteria, options);
        }
        else {
            return model.findOne(criteria, options);
        }
    }
    findOrCreate(table, criteria, creation, options) {
        const model = this.getModel(table);
        return model.findOrCreate(criteria, creation, options);
    }
    update(table, criteria, data, options) {
        const model = this.getModel(table);
        return model.update(criteria, data, options);
    }
    updateOrCreate(table, criteria, data, options) {
        const model = this.getModel(table);
        return model.updateOrCreate(criteria, data, options);
    }
    get(location, criteria, defaultValue) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.get(column, criteria, defaultValue);
    }
    set(location, criteria, value) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.set(column, criteria, value);
    }
    getRaw(location, criteria, defaultValue) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.getRaw(column, criteria, defaultValue);
    }
    setRaw(location, criteria, value) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.setRaw(column, criteria, value);
    }
    incr(location, criteria, amount) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.incr(column, criteria, amount);
    }
    decr(location, criteria, amount, allowNegative) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.decr(column, criteria, amount, allowNegative);
    }
    remove(location, criteria) {
        const model = this.getModel(location);
        return model.remove(criteria);
    }
    clear(location) {
        const model = this.getModel(location);
        return model.clear();
    }
    count(location, criteria, options) {
        if (location == null && criteria == null && options == null) {
            const query = this.knex('sqlite_master')
                .whereNot('name', 'sqlite_sequence')
                .where({ type: 'table' })
                .count('* as count');
            return runQuery(this, query, true)
                .then(([{ count }]) => count);
        }
        const [table, column] = defaultTo(location, '').split('.', 2);
        const model = this.getModel(table);
        return column
            ? model.countIn(column, criteria, options)
            : model.count(criteria, options);
    }
    min(location, criteria, options) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.min(column, criteria, options);
    }
    max(location, criteria, options) {
        const [table, column] = location.split('.', 2);
        const model = this.getModel(table);
        return model.max(column, criteria, options);
    }
}
export { default as Model } from './model';
export * from './types';
export * from './plugins';
export const connect = (path, options) => new Trilogy(path, options);
